from unimeth.extract.patcher import Patcher
from unimeth.config import tokenizer
import numpy as np

class ReadProcessor:

    def __init__(self, args, negative_thres=0, positive_thres=100):
        self.mode = args.mode
        self.detect_CpG, self.detect_CHG, self.detect_CHH, self.detect_m6A = args.cpg, args.chg, args.chh, args.m6A
        self.pore_type, self.frequency, self.dorado_version = args.pore_type, args.frequency, args.dorado_version
        self.negative_thres, self.positive_thres = negative_thres, positive_thres
        self.patcher = Patcher(args)
        
    def weight_bis(self, num):
        return 10 * ((num - 50) / 50) ** 2

    def get_norm_params(self, feature):
        shift_dacs_to_pa, scale_dacs_to_pa = feature['shift_dacs_to_pa'], feature['scale_dacs_to_pa']
        shift_pa_to_norm, scale_pa_to_norm = feature['shift_pa_to_norm'], feature['scale_pa_to_norm']
        if self.frequency == '5khz':
            if self.dorado_version <= 0.71:
                shift = 1-shift_pa_to_norm
                scale = 1/scale_pa_to_norm
            else:
                shift = shift_pa_to_norm
                scale = scale_pa_to_norm
        else:
            shift = (shift_pa_to_norm / scale_dacs_to_pa) - shift_dacs_to_pa
            scale = scale_pa_to_norm / scale_dacs_to_pa
        return shift, scale
    
    def get_methy_type(self, seq, pos):
        if seq[pos] == 'C':
            if pos + 1 < len(seq) and seq[pos+1] == 'G':
                return '[CpG]' if self.detect_CpG else None
            elif pos + 2 < len(seq) and seq[pos+2] == 'G':
                return '[CHG]' if self.detect_CHG else None
            elif pos + 2 < len(seq):
                return '[CHH]' if self.detect_CHH else None
        elif seq[pos] == 'A':
            return '[m6A]' if self.detect_m6A else None

    def add_tokens(self, bases, signal_event, pred_pos, ref_pos=None, bis_label=None):
        meta_data = []
        for i in reversed(range(len(pred_pos))):
            bis_l = -1 if bis_label is None else bis_label[i]
            pos_r = None if ref_pos is None else ref_pos[i]
            pos_p = pred_pos[i]
            methy_type = self.get_methy_type(bases, pos_p)
            
            if methy_type is not None:
                bases.insert(pos_p+1, methy_type)
                signal_event.insert(pos_p+1, [])
                meta_data.append({
                    'ref_pos': pos_r, 
                    'labels': bis_l, 
                    'read_pos': pos_p,
                })     
        meta_data = meta_data[::-1]
        data_dict = {
            'bases': bases,
            'signal_event': signal_event,
        }
        return data_dict, meta_data

    def get_inference_datasets(self, feature):
        new_feature, meta_data = self.add_tokens(
            bases=feature['bases'], 
            signal_event=feature['signal_event'],
            pred_pos=feature['pred_pos'],
            ref_pos=feature['ref_pos'],
            bis_label=feature['bis_label']
        )
        
        shift, scale = self.get_norm_params(feature)
        patch_generator = self.patcher.get_patch_generator(
            mode=self.mode,
            bases=new_feature['bases'], 
            signal_event=new_feature['signal_event'],
            shift=shift,
            scale=scale,
            meta_data=meta_data,
        )
        for patch in patch_generator:
            if len(patch['patch_pos']) == 0:
                continue

            inference_data = {
                'read_id': feature['read_id'],
                'chr': feature['chr'],
                'strand': feature['strand'],
                'signals': patch['signals'],
                'signal_pos': patch['signal_pos'],
                'decoder_input_ids': patch['decoder_input_ids'],
                'patch_pos': patch['patch_pos'],
                'read_pos': patch['read_pos'],
                'ref_pos': patch['ref_pos'],
                'labels': patch['labels']
            }
            
            yield inference_data

    def get_finetune_datasets(self, feature):
        read_label = feature['bis_label']
        new_feature, meta_data = self.add_tokens(
            bases=feature['bases'], 
            signal_event=feature['signal_event'],
            pred_pos=feature['pred_pos'],
            bis_label=read_label
        )
        
        shift, scale = self.get_norm_params(feature)
        patch_generator = self.patcher.get_patch_generator(
            mode=self.mode,
            bases=new_feature['bases'], 
            signal_event=new_feature['signal_event'],
            shift=shift,
            scale=scale,
            meta_data=meta_data,
        )
        for patch in patch_generator:
            label_tokens, weight = [], []
            for x in patch['labels']:
                if x == -1:
                    label_tokens.append(tokenizer['[MASK]'])
                    weight.append(0)
                elif x <= self.negative_thres:
                    label_tokens.append(tokenizer['-'])
                    weight.append(self.weight_bis(x))
                elif x >= self.positive_thres:
                    label_tokens.append(tokenizer['+'])
                    weight.append(self.weight_bis(x))
                else:
                    label_tokens.append(tokenizer['[MASK]'])
                    weight.append(0)
            if (np.array(label_tokens) == tokenizer['[MASK]']).all():
                continue
            train_data = {
                'signals': patch['signals'],
                'signal_pos': patch['signal_pos'],
                'decoder_input_ids': patch['decoder_input_ids'],
                'patch_pos': patch['patch_pos'],
                'labels': label_tokens,
                'weight': weight,
            }
            yield train_data
        
    def get_pretrain_datasets(self, feature):
        shift, scale = self.get_norm_params(feature)
        patch_generator = self.patcher.get_patch_generator(
            mode=self.mode,
            bases=list(feature['bases']), 
            signal_event=feature['signal_event'],
            shift=shift,
            scale=scale,
        )
        for patch in patch_generator:
            train_data = {
                'signals': patch['signals'],
                'signal_pos': patch['signal_pos'],
                'decoder_input_ids': patch['decoder_input_ids']
            }
            yield train_data
    
    def get_datasets(self, feature):
        if self.mode == 'pretrain':
            yield from self.get_pretrain_datasets(feature)
        elif self.mode == 'finetune':
            yield from self.get_finetune_datasets(feature)
        elif self.mode == 'inference':
            yield from self.get_inference_datasets(feature)


'''
    Given pod5 file and bam file, extract read features
'''

import pod5 as p5
from unimeth.util import local_print, state
from torch.utils.data import get_worker_info
from unimeth.extract.feature import Extractor_raw
from unimeth.extract.bam import Read_indexed_bam
from tqdm import tqdm

class Reader_raw:
    def __init__(self, pod5_dir, bam_file:Read_indexed_bam, args):
        self.subset_name = pod5_dir.split('/')[-1]
        self.pod5_file = p5.DatasetReader(pod5_dir, recursive=True, index=True)
        self.bam_file = bam_file
        self.read_ids = self.get_read_ids()
        self.extractor = Extractor_raw(args)

    def get_read_ids(self):
        read_ids = []
        for x in self.pod5_file.read_ids:
            read_ids.append(x)
        return read_ids
    
    def get_features(self):
        worker_info = get_worker_info()
        if worker_info is not None:
            pid, num_workers = worker_info.id, worker_info.num_workers
        else:
            pid, num_workers = 0, 1
        if pid == 0:
            local_print(self.subset_name)
        for i, read_id in enumerate(tqdm(self.read_ids, desc=self.subset_name, disable=(pid != 0 or not state.is_local_main_process))):
            if i % num_workers != pid:
                continue    
            pod5_read = self.pod5_file.get_read(read_id)
            bam_reads = self.bam_file.get_read_by_id(read_id)
            for bam_read in bam_reads:
                if bam_read is None:
                    continue
                feature = self.extractor.get_feature(bam_read, pod5_read)
                if feature is None:
                    continue
                yield feature



import os
import pysam
import pickle
from unimeth.util import local_print

class Read_indexed_bam:
    def __init__(self, bam_dir, force_write = False):
        self.bam_file = pysam.AlignmentFile(bam_dir, "rb", check_sq=False)
        filename = bam_dir.split('/')[-1]
        os.makedirs('data_cache/bam_index', exist_ok = True)
        bam_index_file = f'data_cache/bam_index/{filename}.index'
        if force_write or not os.path.exists(bam_index_file):
            self.bam_index = self._build_bam_index(bam_index_file)
        else:
            self.bam_index = self._load_index(bam_index_file)
    
    def _load_index(self, bam_index_file):
        with open(bam_index_file, 'rb') as fr:
            bam_index = pickle.load(fr)
#        print(f'Load index from {bam_index_file}')
        return bam_index

    def _build_bam_index(self, bam_index_file):
        local_print('Building bam index...')
        bam_index = {}
        read_ptr = self.bam_file.tell()
        for bam_read in self.bam_file:
            if bam_read.is_supplementary or bam_read.is_secondary:
                read_ptr = self.bam_file.tell()
                continue
            if bam_read.has_tag('pi'):
                read_id = bam_read.get_tag("pi")
            else:
                read_id = bam_read.query_name
            if read_id not in bam_index:
                bam_index[read_id] = []    
            bam_index[read_id].append(read_ptr)
            read_ptr = self.bam_file.tell()
        with open(bam_index_file, 'wb') as fw:
            pickle.dump(bam_index, fw)
        local_print(f'Bam index built and saved in {bam_index_file}. Size: {len(bam_index)}')
        return bam_index
    
    def get_read_by_id(self, read_id):
        bam_reads = []
        if read_id not in self.bam_index:
#            print(read_id, "Not in bam_index")
            return bam_reads
        read_ptrs = self.bam_index[read_id]    
        for read_ptr in read_ptrs:
            self.bam_file.seek(read_ptr)
            bam_read = next(self.bam_file)
            bam_reads.append(bam_read)
        return bam_reads


import numpy as np
from unimeth.config import methy_types, tokenizer

class Patcher:
    def __init__(self, args):
        self.overlap, self.chunk_size = args.overlap, args.chunk_size
        self.mode = args.mode
        self.use_prune = args.use_prune

    def prune(self, chunk_bases, ignore_left, ignore_right):
        first, last = len(chunk_bases), 0 
        for j in range(ignore_left, len(chunk_bases)-(0 if ignore_right == 0 else self.overlap)):
            if chunk_bases[j] in ['-', '+', '[MASK]']:
                first = min(first, j)
                last = max(last, j)
        if last == 0:
            return 0, 0
        first = max(first - self.overlap, 0)
        last = min(last + self.overlap + 1, len(chunk_bases))
        return first, last

    def get_patch_generator(self, mode, bases, signal_event, shift, scale, meta_data=[]):
        seq_len = len(bases)
        overlap = self.overlap
        chunk_size = self.chunk_size
        gap = chunk_size - overlap * 2
        meta_idx = 0
        for i in range(0, seq_len-overlap*2, gap):
            # sequence processing
            left, right = i, i + chunk_size
            chunk_bases = bases[left: right]
            chunk_signal_event = signal_event[left : right]
            chunk_bases = np.array([tokenizer[atcg] for atcg in chunk_bases], dtype=np.int8)
            ignore_left = overlap if i > 0 else 0
            ignore_right = overlap + 1 if i + gap < seq_len-overlap*2 else 0

            methys = np.array([tokenizer[atcg] for atcg in methy_types])
            result = np.where(np.isin(chunk_bases, methys))[0]

            meta_input = {
                'patch_pos': [],
                'read_pos': [],
                'ref_pos': [],
                'labels': [],
            }
            first, last = len(chunk_bases), 0 
            for patch_pos in result:
                if patch_pos < ignore_left or patch_pos > len(chunk_bases) - ignore_right:
                    continue
                first = min(first, patch_pos)
                last = max(last, patch_pos)
                meta = meta_data[meta_idx]
                meta_input['patch_pos'].append(patch_pos)
                if 'read_pos' in meta:
                    meta_input['read_pos'].append(meta['read_pos']) 
                if 'ref_pos' in meta:
                    meta_input['ref_pos'].append(meta['ref_pos'])
                if 'labels' in meta:
                    meta_input['labels'].append(meta['labels'])
                meta_idx += 1
            if self.use_prune and last != 0:
                left = max(first - self.overlap, 0)
                right = min(last + self.overlap + 1, len(chunk_bases))
                chunk_bases = chunk_bases[left:right]
                chunk_signal_event = chunk_signal_event[left:right]
                meta_input['patch_pos'] = np.array(meta_input['patch_pos']) - left

            # get signal pos
            signal_pos = [[i] * len(sublist) for i, sublist in enumerate(chunk_signal_event)]
            signal_pos = sum(signal_pos, [])
            signal_pos = np.array(signal_pos[::4], dtype=np.int32)
            assert len(chunk_bases) == len(chunk_signal_event)

            chunk_bases = np.append(chunk_bases, tokenizer['[END]'])
            chunk_signal = np.hstack(chunk_signal_event)
            chunk_signal = chunk_signal = (chunk_signal - shift) / scale
            if mode == 'pretrain':
                data_dict = {
                    'signals': chunk_signal,
                    'signal_pos': signal_pos,
                    'decoder_input_ids': chunk_bases,
                }
            else: 
                data_dict = {
                    'signals': chunk_signal,
                    'signal_pos': signal_pos,
                    'decoder_input_ids': chunk_bases,
                    'patch_pos': meta_input['patch_pos'],
                    'labels': meta_input['labels'],
                    'read_pos': meta_input['read_pos'],
                    'ref_pos': meta_input['ref_pos'],
                }
            yield data_dict
        assert meta_idx == len(meta_data), print(meta_idx, len(meta_data))


import os
import torch
from torch.utils.data import IterableDataset, Dataset
from unimeth.extract.reader import Reader_raw
from unimeth.extract.processors import ReadProcessor
from unimeth.extract.bam import Read_indexed_bam
from unimeth.util import local_print
from itertools import zip_longest

class MultiDataset(IterableDataset):
    def __init__(self, generators):
        self.generators = generators
    def __iter__(self):
        for binned_data in zip_longest(*self.generators, fillvalue=None):
            for data in binned_data:
                yield data
    
class RawDataset(IterableDataset):
    def __init__(self, pod5_dir, bam_dir, args, negative_thres=0, positive_thres=100):
        if os.path.isdir(pod5_dir):
            subsets = os.listdir(pod5_dir)
            self.pod5_dirs = [os.path.join(pod5_dir, x) for x in subsets]
        else:
            self.pod5_dirs = [pod5_dir]
        self.bam_dir = bam_dir
        self.args = args
        self.read_processor = ReadProcessor(self.args, negative_thres, positive_thres)
        self.binning = Binning(args)

    def __iter__(self):
        bam_file = Read_indexed_bam(self.bam_dir, force_write=True)
        for pod5_dir in self.pod5_dirs:
            reader = Reader_raw(pod5_dir=pod5_dir, bam_file=bam_file, args=self.args)
            for feature in reader.get_features():
                for dataset in self.read_processor.get_datasets(feature):
                    for binned_data in self.binning.get_data(dataset):
                        yield binned_data
        yield from self.binning.flush()

class RawValDataset(Dataset):
    def __init__(self, pod5_dirs, bam_dirs, args):
        if len(pod5_dirs) != len(bam_dirs):
            print('Need same number of pod5 files and bam files')
        self.datas = []
        total_max = 5000
        per_max = total_max // len(pod5_dirs)
        for i in range(len(pod5_dirs)):
            dataset = RawDataset(pod5_dirs[i], bam_dirs[i], args)    
            for num, x in enumerate(dataset):
                self.datas.append(x)
                if num > per_max:
                    break
        local_print(f'Validation data: {len(self.datas)}')

    def __getitem__(self, idx):
        return self.datas[idx]

    def __len__(self):
        return len(self.datas)


def collate_fn(mode, datas):
    signals = [torch.tensor(sample["signals"], dtype=torch.float) for sample in datas]    
    padded_signals = torch.nn.utils.rnn.pad_sequence(signals, batch_first=True, padding_value=0)

    decoder_inputs = [torch.tensor(sample["decoder_input_ids"], dtype=torch.long) for sample in datas]
    padded_decoder_inputs = torch.nn.utils.rnn.pad_sequence(decoder_inputs, batch_first=True, padding_value=-100)

    signal_pos = [torch.tensor(sample["signal_pos"], dtype=torch.int) for sample in datas]
    padded_signal_pos = torch.nn.utils.rnn.pad_sequence(signal_pos, batch_first=True, padding_value=256)

    encoder_masks = [torch.ones(((len(sample["signals"]) - 1) // 4) + 1, dtype=torch.long) for sample in datas]
    padded_encoder_masks = torch.nn.utils.rnn.pad_sequence(encoder_masks, batch_first=True, padding_value=0)

    if mode == 'inference':
        ret = {
            'signals': padded_signals,
            'encoder_mask': padded_encoder_masks,
            'decoder_input_ids': padded_decoder_inputs,
            'signal_pos': padded_signal_pos,

            'patch_pos': [sample['patch_pos'] for sample in datas],
            'read_pos': [sample['read_pos'] for sample in datas],
            'ref_pos': [sample['ref_pos'] for sample in datas],
            'labels': [sample['labels'] for sample in datas],
            'read_id': [sample['read_id'] for sample in datas],
            'chr': [sample['chr'] for sample in datas],
            'strand': [sample['strand'] for sample in datas],
        }
    elif mode == 'finetune':
        padded_label = torch.ones_like(padded_decoder_inputs) * -100
        padded_weight = torch.ones_like(padded_decoder_inputs)
        for i, sample in enumerate(datas):
            patch_pos, label, weight = sample['patch_pos'], sample['labels'], sample['weight']
            for j in range(len(patch_pos)):
                pos, label_pos, weight_pos = patch_pos[j], label[j], weight[j]
                padded_label[i, pos] = label_pos
                padded_weight[i, pos] = weight_pos
        ret = {
            'signals': padded_signals,
            'encoder_mask': padded_encoder_masks,
            'decoder_input_ids': padded_decoder_inputs,
            'signal_pos': padded_signal_pos,
            'labels': padded_label,
            'weights': padded_weight,
        }
    elif mode == 'pretrain':
        ret = {
            'signals': padded_signals,
            'encoder_mask': padded_encoder_masks,
            'labels': padded_decoder_inputs,
            'signal_pos': padded_signal_pos,
        }
    return ret

import functools
collate_fn_pretrain = functools.partial(collate_fn, 'pretrain')
collate_fn_finetune = functools.partial(collate_fn, 'finetune')
collate_fn_inference = functools.partial(collate_fn, 'inference')


class Binning:

    def __init__(self, args):
        self.num_bins = args.num_bins
        self.bin_size, self.max_bin_length = args.bin_size, args.max_bin_length
        self.bins = [[] for _ in range(self.num_bins)]

    def get_data(self, dataset):
        l = len(dataset['signals'])
        if l < 50 or l >= self.bin_size * self.num_bins:
            return
        bin_id = l // self.bin_size
        self.bins[bin_id].append(dataset)
        if len(self.bins[bin_id]) >= self.max_bin_length:
            for x in self.bins[bin_id]:
                yield x
            self.bins[bin_id] = []

    def flush(self):
        for bin in self.bins:
            for x in bin:
                yield x



'''
    Extract features from given (bam_read, pod5_read) pair
'''

import numpy as np

def get_chr_list(chr_str):
    if '|' not in chr_str:
        print('error, default as using all chr')
        return 'exclude', set()
    parts = chr_str.split('|')
    if parts[0] != '' and parts[1] != '':
        print('error, default as using all chr')
        return 'exclude', set()
    elif parts[0] == '' and parts[1] == '':
        chr_mode, chr_list = 'exclude', set()
    elif parts[0] != '':
        chr_mode, chr_list = 'exclude', set(parts[0].split(','))
    else:
        chr_mode, chr_list = 'include', set(parts[1].split(','))
    return chr_mode, chr_list

def get_ref_pos(seq, pred_pos, aligned_pairs, is_reverse):
    '''
        Given read sequence, aligned_pairs and if it is reverse
        Output reference position, len(seq) = len(ref_pos)
    '''
    ref_pos = []
    if len(pred_pos) == 0 or len(aligned_pairs) == 0:
        return [-1] * len(pred_pos)
    idx = 0
    for ps, pr in aligned_pairs:
        if ps is None:
            continue
        ps = len(seq) - ps - 1 if is_reverse else ps
        if ps == pred_pos[idx]:
            ref_pos.append(pr if pr is not None else -1)
            idx += 1
        elif ps > pred_pos[idx]:
            ref_pos.append(-1)
            idx += 1
        if idx == len(pred_pos):
            break
    assert len(ref_pos) == len(pred_pos), f'{len(ref_pos)} and {len(pred_pos)}'
    return ref_pos


basepairs = {
    "A": "T",
    "C": "G",
    "G": "C",
    "T": "A",
}

def alphabet(letter):
    if letter in basepairs.keys():
        return basepairs[letter]
    return "N"

def complement_seq(base_seq):
    rbase_seq = base_seq[::-1]
    comseq = ""
    try:
        comseq = "".join([alphabet(x) for x in rbase_seq])
    except Exception:
        print("something wrong in the dna/rna sequence.")
    return comseq

def align_to_ref(bam_read, seq, signal_event, aligned_pairs, mod_dict):
    try:
        ref_seq = bam_read.get_reference_sequence().upper()
    except:
        return None
    if bam_read.is_reverse:
        ref_seq = complement_seq(ref_seq)
    ref_start = bam_read.reference_start
    first_index = next((i for i in range(len(aligned_pairs)) if aligned_pairs[i][1] is not None), len(aligned_pairs))
    last_index = next((i for i in range(len(aligned_pairs)-1, 0, -1) if aligned_pairs[i][1] is not None), len(aligned_pairs))
    new_ref_seq, new_ref_signal, new_ref_pos = [], [], []
    new_mod_dict = {}
    for i, (ps, pr) in enumerate(aligned_pairs):
        if ps is not None and bam_read.is_reverse:
            ps = len(seq) - ps - 1
        signal = [] if ps is None or signal_event is None else signal_event[ps]
        if i < first_index or i > last_index:
            if ps is not None:
                if ps in mod_dict:
                    new_mod_dict[len(new_ref_seq)] = mod_dict[ps]
                new_ref_seq.append(seq[ps])
                new_ref_signal.append(signal)
                new_ref_pos.append(pr)
        else:
            if pr is not None:
                if bam_read.is_reverse:
                    ref_base = ref_seq[len(ref_seq) - (pr - ref_start) - 1] 
                else:
                    ref_base = ref_seq[pr - ref_start]
                if ps in mod_dict:
                    new_mod_dict[len(new_ref_seq)] = mod_dict[ps]
                new_ref_seq.append(ref_base)
                new_ref_signal.append(signal)
                new_ref_pos.append(pr)
            elif len(new_ref_signal) > 0:
                new_ref_signal[-1] = np.append(new_ref_signal[-1], signal)
    seq = ''.join(new_ref_seq)
    signal_event = new_ref_signal
    if bam_read.is_reverse:
        aligned_pairs = [(len(new_ref_seq) - 1 - i, new_ref_pos[i]) for i in range(len(new_ref_seq))]
    else:
        aligned_pairs = [(i, new_ref_pos[i]) for i in range(len(new_ref_seq))]
    return seq, signal_event, aligned_pairs, new_mod_dict

class Pred_pos_extractor:
    def __init__(self, args):
        self.detect_CpG = args.cpg
        self.detect_CHG = args.chg
        self.detect_CHH = args.chh
        self.detect_m6A = args.m6A

    def get_methy_pos(self, seq):
        pred_pos = []
        for i in range(len(seq)):
            if seq[i] == 'C':
                if i + 1 < len(seq) and seq[i+1] == 'G' and self.detect_CpG:
                    pred_pos.append(i)
                elif i + 2 < len(seq) and seq[i+2] == 'G' and self.detect_CHG:
                    pred_pos.append(i)
                elif i + 2 < len(seq) and self.detect_CHH:
                    pred_pos.append(i)
            elif seq[i] == 'A':
                if self.detect_m6A:
                    pred_pos.append(i)
        return pred_pos

class Extractor_raw:
    def __init__(self, args):
        self.mapq_thres = args.mapq_thres
        self.pred_pos_extractor = Pred_pos_extractor(args)
        if args.m6A == 1:
            self.detect_mod = ('A',0,'a')
        elif args.cpg or args.chg or args.chh:
            self.detect_mod = ('C',0,'m')
        else:
            self.detect_mod = None
        self.align_ref = (args.pore_type == 'R9.4.1')
        self.chr_mode, self.chr_list = get_chr_list(args.chr)
    
    #@jit(nopython=True)
    def get_signal(self, signal, move):
        stride, start, move_table=move
        move_index=np.where(move_table)[0]
        event=[]
        for i in range(len(move_index) - 1):
            prev=move_index[i]*stride+start
            sig_end=move_index[i+1]*stride+start
            event.append(signal[prev:sig_end])
        prev=move_index[len(move_index) - 1]*stride+start
        event.append(signal[prev:])
        return event

    def get_move(self, bam_read):
        read_dict = bam_read.to_dict()
        tags={x.split(':')[0]:x for x in read_dict.pop('tags')}
        start=int(tags['ts'].split(':')[-1])
        if bam_read.has_tag('sp'):
            start += bam_read.get_tag('sp')
        mv=tags['mv'].split(',')
        stride=int(mv[1])
        move_table=np.fromiter(mv[2:], dtype=np.int8)
        move=(stride, start, move_table)
        return move

    def get_feature(self, bam_read, pod5_read):
        mapq = bam_read.mapping_quality
        if mapq < self.mapq_thres:
            print(f'MapQ < {self.mapq_thres}')
            return None
        chr = bam_read.reference_name
        if chr is None:
            return None
        elif self.chr_mode == 'exclude' and chr in self.chr_list:
            return None
        elif self.chr_mode == 'include' and chr not in self.chr_list:
            return None

        seq = bam_read.get_forward_sequence().upper()

        signal = pod5_read.signal
        shift_dacs_to_pa=pod5_read.calibration.offset
        scale_dacs_to_pa=pod5_read.calibration.scale
        shift_pa_to_norm = bam_read.get_tag("sm")
        scale_pa_to_norm = bam_read.get_tag("sd")
        
        move = self.get_move(bam_read)
        signal_event = self.get_signal(signal, move)
        aligned_pairs = bam_read.get_aligned_pairs()
        mod = bam_read.modified_bases_forward
        mod_type = mod[self.detect_mod] if self.detect_mod in mod else []
        mod_dict = {}
        for pos, label in mod_type:
            mod_dict[pos] = label
        
        if bam_read.is_reverse:
            aligned_pairs = aligned_pairs[::-1]
        if self.align_ref:
            seq, signal_event, aligned_pairs, mod_dict = align_to_ref(bam_read, seq, signal_event, aligned_pairs, mod_dict)

        pred_pos = self.pred_pos_extractor.get_methy_pos(seq)
        ref_pos = get_ref_pos(seq, pred_pos, aligned_pairs, bam_read.is_reverse)

        bis_label = []
        for pos in pred_pos:
            if pos in mod_dict:
                bis_label.append(int(round(mod_dict[pos] / 255 * 100)))
            else:
                bis_label.append(-1)

        output = {
            'read_id': bam_read.query_name,
            'chr': chr,
            'reference_start': bam_read.reference_start,
            'bases': list(seq),
            'signal_event': signal_event,
            'pred_pos': pred_pos,
            'bis_label': bis_label,
            'mapQ': mapq,
            'shift_dacs_to_pa': shift_dacs_to_pa,
            'scale_dacs_to_pa': scale_dacs_to_pa,
            'shift_pa_to_norm': shift_pa_to_norm,
            'scale_pa_to_norm': scale_pa_to_norm,
            'ref_pos': ref_pos,
            'strand': '-' if bam_read.is_reverse else '+',
        }
        return output





from torch.utils.data import DataLoader
from unimeth.model.data import collate_fn_inference, RawDataset
import torch
from accelerate import Accelerator
from accelerate.utils import DataLoaderConfiguration
from unimeth.model.model import Basecaller
from unimeth.config import defaultconfig, vocab
from unimeth.util import local_print
from pathlib import Path

def get_result(decoder_input_ids, logits, patch_pos): 
    num_data = logits.shape[0]
    preds, methy = [], []
    for i in range(num_data):
        logits_patch = logits[i]
        pos = patch_pos[i]
        site_logits = logits_patch[pos]
        preds_patch = torch.exp(site_logits[:, 10]) / torch.sum(torch.exp(site_logits[:, [10, 11]]), axis = 1)
        methy_patch = decoder_input_ids[i][pos]
        preds.append(preds_patch)
        methy.append(methy_patch)
    return torch.hstack(preds), torch.hstack(methy)


def inference(args):
    dataLoaderConfiguration = DataLoaderConfiguration(
        dispatch_batches=False
    )
    accelerator = Accelerator(
        dataloader_config=dataLoaderConfiguration
    )
    dataset = RawDataset(pod5_dir=args.pod5_dir,bam_dir=args.bam_dir,args=args)

    # 创建 DataLoader
    dataloader = DataLoader(
        dataset,
        collate_fn=collate_fn_inference,
        batch_size=args.batch_size,  
        num_workers=args.num_workers,  
        pin_memory=True,  
        prefetch_factor=2  
    )
    dataloader = accelerator.prepare(dataloader)
    model = Basecaller(mode='inference')
    model.load_state_dict(torch.load(args.model_dir), strict=True)
    model.eval()
    model = accelerator.prepare(model)
    
    file_path = Path(args.out_dir)
    file_path.parent.mkdir(parents=True, exist_ok=True)

    with torch.no_grad(), torch.cuda.amp.autocast(), open(args.out_dir, 'w') as fw:
        for num_batch, batch in enumerate(dataloader):
            if args.limit is not None and num_batch > args.limit:
                break
            logits = model(
                signals = batch['signals'],
                encoder_mask = batch['encoder_mask'],
                decoder_input_ids = batch['decoder_input_ids'],
                signal_pos = batch['signal_pos'],
            )
            logits = accelerator.pad_across_processes(logits, dim=1, pad_index=0)
            decoder_input_ids = accelerator.pad_across_processes(batch['decoder_input_ids'], dim=1, pad_index=0)
            logits = accelerator.gather(logits).cpu()
            decoder_input_ids = accelerator.gather(decoder_input_ids).cpu()
            patch_pos = accelerator.gather_for_metrics(batch['patch_pos'])
            read_id = accelerator.gather_for_metrics(batch['read_id'])
            chr = accelerator.gather_for_metrics(batch['chr'])
            strand = accelerator.gather_for_metrics(batch['strand'])
            ref_pos = accelerator.gather_for_metrics(batch['ref_pos'])
            read_pos = accelerator.gather_for_metrics(batch['read_pos'])
            labels = accelerator.gather_for_metrics(batch['labels'])

            preds, methy = get_result(decoder_input_ids, logits, patch_pos)
            idx = 0
            for i in range(len(read_id)):
                patch_pos_i = patch_pos[i]
                for j in range(len(patch_pos_i)):
                    prob1 = preds[idx].item()
                    prob0 = 1 - prob1
                    probbool = int(prob1 > 0.5)
                    line = [chr[i], str(ref_pos[i][j]), strand[i], str(labels[i][j]), read_id[i], str(read_pos[i][j]), vocab[methy[idx].item()], str(prob0), str(prob1), str(probbool), '.']
                    try:
                        line = '\t'.join(line)
                    except:
                        print(line)
                    fw.write(line + '\n')
                    idx += 1
            assert idx == len(preds)

def main(args):
    for key, value in defaultconfig.items():
        if not hasattr(args, key) or getattr(args, key) is None:
            setattr(args, key, value)
    args.mode='inference'
    local_print(args)
    inference(args)


from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser

__version__ = '0.0.1'

def start():
    parser = ArgumentParser(
        'unimeth', 
        description='Unimeth is a unified deep learning framework for accurate and efficient detection of DNA methylation (5mC, 6mA) from Oxford Nanopore sequencing data. Built on a transformer-based architecture, Unimeth supports multiple sequencing chemistries (R9.4.1, R10.4.1 4kHz/5kHz), handles both plant and mammalian genomes, and achieves state-of-the-art performance across diverse genomic contexts.',
        formatter_class=ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        '-v', '--version', 
        action='version', 
        version=f'unimeth {__version__}'
    )
    
    input_group = parser.add_argument_group('Input/Output options')
    input_group.add_argument(
        '--pod5_dir', type=str, required=True,
        help='Directory containing POD5 files (raw signal data)'
    )
    input_group.add_argument(
        '--bam_dir', type=str, required=True,
        help='Directory containing BAM files (aligned sequencing data)'
    )
    input_group.add_argument(
        '--model_dir', type=str, required=True,
        help='Path to the trained model file (.pt)'
    )
    input_group.add_argument(
        '--out_dir', type=str, required=True,
        help='Output file path for methylation results'
    )
    
    chr_group = parser.add_argument_group('Chromosome filtering options')
    chr_group.add_argument(
        '--chr', type=str, default='|',
        help=
        '''
        Chromosome selection in format "exclude_chr|include_chr". 
        Examples:
        - "|" : process all chromosomes (default)
        - "chrM,chrY|" : exclude chrM and chrY
        - "|chr1,chr2,chr3" : only include chr1, chr2, chr3
        - "chrM|chr1,chr2" : invalid, will use all chromosomes
        '''
    )
    
    control_group = parser.add_argument_group('Runtime control options')
    control_group.add_argument(
        '--limit', type=int, default=None,
        help='Limit number of batches to process (for testing)'
    )
    control_group.add_argument(
        '--batch_size', type=int, default=512,
        help='Batch size for inference'
    )
    
    meth_group = parser.add_argument_group('Methylation type options')
    meth_group.add_argument(
        '--cpg', type=int, default=0, choices=[0, 1],
        help='Detect CpG methylation sites (0: disable, 1: enable)'
    )
    meth_group.add_argument(
        '--chg', type=int, default=0, choices=[0, 1],
        help='Detect CHG methylation sites (0: disable, 1: enable)'
    )
    meth_group.add_argument(
        '--chh', type=int, default=0, choices=[0, 1],
        help='Detect CHH methylation sites (0: disable, 1: enable)'
    )
    meth_group.add_argument(
        '--m6A', type=int, default=0, choices=[0, 1],
        help='Detect m6A methylation sites (0: disable, 1: enable)'
    )
    
    # 技术参数
    tech_group = parser.add_argument_group('Technical options')
    tech_group.add_argument(
        '--pore_type', type=str, required=True, choices=['R10.4.1', 'R9.4.1'],
        help='Nanopore pore type'
    )
    tech_group.add_argument(
        '--frequency', type=str, required=True, choices=['4khz', '5khz'],
        help='Sequencing frequency (e.g., "4000" for 4kHz)'
    )
    tech_group.add_argument(
        '--dorado_version', type=float, required=True, choices=[0.71, 0.81],
        help='Dorado basecaller version used for data generation'
    )
    args = parser.parse_args()
    
    from unimeth.inference import main
    main(args)


import torch
from torch import Tensor, nn
from torch.nn import CrossEntropyLoss
from transformers import BartForConditionalGeneration, BartConfig
from transformers.models.bart.modeling_bart import shift_tokens_right
from unimeth.model.module import Processor, PositionalEncoding
from unimeth.config import modelconfig

class Basecaller(nn.Module):
    def __init__(self, mode, plant=False):
        super().__init__()
        self.processor = Processor(hidden_size=modelconfig['d_model'])
        self.encoder_decoder = BartForConditionalGeneration(BartConfig(**modelconfig))
        self.pos_encoder = PositionalEncoding(d_model=modelconfig['d_model'])
        self.entropy_loss = CrossEntropyLoss(reduction='none')
        self.mode = mode
        self.plant=plant
        self.methy_types = [7, 8, 9, 13]
    
    def _forward(
            self, 
            signals: Tensor, 
            encoder_mask: Tensor, 
            signal_pos: Tensor, 
            decoder_input_ids: Tensor=None, 
            labels: Tensor=None, 
            ):
        x = self.processor(signals)
        x = self.pos_encoder(x, signal_pos)
        if self.mode == 'pretrain':
            decoder_mask = None
        else:
            decoder_input_ids = shift_tokens_right(decoder_input_ids, 0, 1)
            batch_size, seq_length = decoder_input_ids.shape
            device = decoder_input_ids.device
            decoder_mask = torch.ones(size=(batch_size, seq_length, seq_length), device=device)        
            
        outputs = self.encoder_decoder(
            inputs_embeds=x, 
            attention_mask=encoder_mask, 
            labels=labels,
            decoder_input_ids=decoder_input_ids,
            decoder_attention_mask=decoder_mask,
        )
        return outputs
    
    def forward(
            self, 
            signals: Tensor, 
            encoder_mask: Tensor, 
            signal_pos: Tensor, 
            decoder_input_ids: Tensor=None, 
            labels: Tensor=None, 
            weights: Tensor=None, 
        ):

        outputs = self._forward(
            signals = signals, 
            encoder_mask = encoder_mask,
            signal_pos = signal_pos,
            decoder_input_ids = decoder_input_ids,
            labels=labels,
        )
        logits, loss = outputs.logits, outputs.loss
        if self.mode == 'pretrain':
            return loss, None
        elif self.mode == 'finetune':
            if self.plant:
               return self.get_unbalance_loss(logits, labels, decoder_input_ids), logits
            else:
               return self.get_weighted_loss(logits, labels, weights), logits
                
        elif self.mode == 'inference':
            return logits
        

    def get_weighted_loss(self, logits, labels, weights):
        label_mask = (labels.view(-1) != -100)
        methy_loss = self.entropy_loss(input=logits.view(-1, logits.size(-1)), target=labels.view(-1))
        weighted_loss = methy_loss * weights.view(-1)
        methy_loss = torch.mean(weighted_loss[label_mask])
        return methy_loss

    
    def get_unbalance_loss(self, logits, labels, decoder_input_ids):
        total_loss, total_items, new_loss = 0, 0, 0
        for methy_token in self.methy_types:
            mask1 = decoder_input_ids == methy_token
            for methy_label in [11, 10]:
                mask2 = (labels == methy_label)
                mask = mask1 & mask2
                if True not in mask:
                    continue
                subloss = self.entropy_loss(logits[mask], labels[mask]).mean()
                num_items = len(logits[mask])
                total_loss += subloss * num_items
                total_items += num_items
                new_loss += subloss
        return new_loss



import math
import torch
from torch import Tensor, nn


class Processor(nn.Module):
    def __init__(self, hidden_size):
        super().__init__()
        self.process = nn.Sequential(
            nn.Conv1d(in_channels=1, out_channels=hidden_size//2, kernel_size=3, stride=2, padding=1),
            nn.BatchNorm1d(num_features=hidden_size//2),
            nn.GELU(),
            nn.Conv1d(in_channels=hidden_size//2, out_channels=hidden_size, kernel_size=3, stride=2, padding=1),
            nn.BatchNorm1d(num_features=hidden_size),
            nn.GELU()
        )
    
    def forward(self, x: Tensor):
        x = x.reshape(x.size()[0], 1, -1)
        x = self.process(x)
        x = x.permute(0, 2, 1)
        return x

class PositionalEncoding(nn.Module):
    def __init__(self, d_model, max_len=260):
        super(PositionalEncoding, self).__init__()
        encoding = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * -(math.log(10000.0) / d_model))

        encoding[:, 0::2] = torch.sin(position * div_term)  # 偶数维度
        encoding[:, 1::2] = torch.cos(position * div_term)  # 奇数维度
        self.register_buffer('encoding', encoding)

    def forward(self, x, pos):
        return x + self.encoding[pos]

class Threshold_based_loss(nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, logits, pos_ratio):
        n = logits.size(0)
        num_positive = int(torch.round(pos_ratio * n))
        sorted_logits, indices = torch.sort(logits, descending=True)
        pos_samples = sorted_logits[:num_positive]
        neg_samples = sorted_logits[num_positive:]
        if len(pos_samples) > 0:
            pos_loss = -torch.log(pos_samples)
        else:
            pos_loss = torch.tensor([], device=logits.device)
        if len(neg_samples) > 0:
            neg_loss = -torch.log(1 - neg_samples)
        else:
            neg_loss = torch.tensor([], device=logits.device)
        total_loss = torch.concat([pos_loss, neg_loss]).mean()
        return total_loss



from transformers import TrainingArguments, TrainerCallback
from transformers.trainer_callback import TrainerControl, TrainerState
class Mycallback(TrainerCallback):
    def on_save(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        if state.is_local_process_zero:
            print(f'Save model on step {state.global_step}')
        return super().on_save(args, state, control, **kwargs)

    def on_log(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, logs=None, **kwargs):
        logs['steps'] = state.global_step
        return super().on_log(args, state, control, **kwargs)


import os
import time
os.environ['TZ'] = 'Asia/Shanghai'
time.tzset()

import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
warnings.simplefilter(action='ignore', category=UserWarning)

defaultconfig={
    'vocab': ['[PAD]', '[START]', '[END]', 'A', 'G', 'T', 'C', '[CpG]', '[CHG]', '[CHH]', '+', '-', '[MASK]', '[m6A]', '[R10]', '[4khz]', '[5khz]'],
    'num_workers':4,
    'batch_size':512,
    'num_bins': 15,
    'bin_size':500,
    'max_bin_length':512,
    'chunk_size':256,
    'overlap':16,
    'cpg':0,
    'chg':0,
    'chh':0,
    'm6A':0,
    'pore_type': 'r9.4.1',
    'frequency': '4khz',
    'mapq_thres': 0,
    'dorado_version': 0.71,
    'use_prune': 0,
}
methy_types = ['[CpG]', '[CHG]', '[CHH]', '[m6A]']
vocab = defaultconfig['vocab']
tokenizer = {vocab[i]: i for i in range(len(vocab))}
modelconfig = {
    "d_ff": 2048,
    "d_kv": 64,
    "d_model": 384,
    "max_position_embeddings": 4096,
    "dropout_rate": 0.1,
    "feed_forward_proj": "gated-gelu",
    "is_encoder_decoder": True,
    "num_heads": 8,
    "num_layers": 12,
    "pad_token_id": tokenizer['[PAD]'],
    "eos_token_id": tokenizer['[END]'],
    "decoder_start_token_id": tokenizer['[START]'],
    "vocab_size": len(vocab)
}




from unimeth.config import vocab

from accelerate.state import PartialState
state = PartialState()

def token2seq(tokens):
    seq = [vocab[x] for x in tokens if x != -100]
    seq = ''.join(seq)
    return seq

@state.on_local_main_process
def local_print(s):
    print(s)

